@@ -11,23 +11,25 @@
  */
 
 enum intel_options {
+	OPTION_ACCEL_DISABLE,
 	OPTION_ACCEL_METHOD,
+	OPTION_BACKLIGHT,
 	OPTION_DRI,
 	OPTION_VIDEO_KEY,
 	OPTION_COLOR_KEY,
 	OPTION_TILING_2D,
 	OPTION_TILING_FB,
 	OPTION_SWAPBUFFERS_WAIT,
+	OPTION_TRIPLE_BUFFER,
 	OPTION_PREFER_OVERLAY,
 	OPTION_HOTPLUG,
 	OPTION_RELAXED_FENCING,
-#ifdef INTEL_XVMC
+#if defined(XvMCExtension) && defined(ENABLE_XVMC)
 	OPTION_XVMC,
+#define INTEL_XVMC 1
 #endif
 #ifdef USE_SNA
-	OPTION_THROTTLE,
 	OPTION_ZAPHOD,
-	OPTION_DELAYED_FLUSH,
 	OPTION_TEAR_FREE,
 	OPTION_CRTC_PIXMAPS,
 #endif
@@ -37,7 +39,6 @@ enum intel_options {
 	OPTION_DEBUG_FLUSH_CACHES,
 	OPTION_DEBUG_WAIT,
 	OPTION_BUFFER_CACHE,
-	OPTION_TRIPLE_BUFFER,
 #endif
 	NUM_OPTIONS,
 };
